﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Xml.Serialization;
using UniversalPlantViewerBuilderPlugin.Version1;
using UniversalPlantViewerBuilderPluginInternal;
using System.Threading.Tasks;
using System.Diagnostics;

namespace UPVBPluginGraphics
{
    public class Plugin : IPluginVersion1
    {
        private Config _config;
        public int ProcessOrder => 10;
        public string Name => "UPVPluginGraphics";
        

        private Config ReadXmlConfig(IPreExportHook hook)
        {
            var ConfigFileName = "UPVBPluginGraphics.xml";

            //search for config in first session directory
            var firstSession = hook.BuilderElementAccessor.Sessions.First();
            var configPath = Path.Combine(Path.GetDirectoryName(firstSession.SessionFile), ConfigFileName);
            string path;
            if (File.Exists(configPath))
            {
                //use session config
                path = configPath;
            }
            else
            {
                //fallback: use plugin folder
                path = Path.GetDirectoryName(Uri.UnescapeDataString(new UriBuilder(Assembly.GetExecutingAssembly().CodeBase).Path));
                if (path == null)
                {
                    throw new InvalidOperationException();
                }
                path = Path.Combine(path, ConfigFileName);
            }

            hook.BuilderContext.Log($"{Name}: reading config from '{path}'");

            var reader = new XmlSerializer(typeof(Config));
            using (var file = new StreamReader(path))
            {
                return (Config)reader.Deserialize(file);
            }
        }

        public void BeforeExport(IPreExportHook hook)
        {
            hook.BuilderContext.Log($"{Name}: started...");

            _config = ReadXmlConfig(hook);

            try
            {
                CreateAspects(hook);
                var objectsToProcess = IdentifyObjects(hook);
                GenerateGeometry(hook, objectsToProcess);
                WriteGeometry(hook, objectsToProcess);
            }
            catch (Exception ex)
            {
                hook.BuilderContext.LogError(ex.Message, ex);
            }

            hook.BuilderContext.Log($"{Name}: finished.");
        }

        private void CreateAspects(IPreExportHook hook)
        {
            hook.BuilderContext.Log($"{Name}: Creating Aspects..");
            foreach (var definition in _config.Definitions)
            {
                if (!hook.BuilderGraphicAccessor.GetAspects().Contains(definition.AspectName))
                {
                    hook.BuilderGraphicAccessor.CreateAspect(definition.AspectName);
                }
            }
            hook.BuilderContext.Log($"{Name}: Creating Aspects done.");
        }

        private List<ObjectToProcess> IdentifyObjects(IPreExportHook hook)
        {
            var ret = new List<ObjectToProcess>();
            hook.BuilderContext.Log($"{Name}: Identifiying objects..");
            foreach (var element in hook.BuilderElementAccessor.IterateElements())
            {
                foreach (var definition in _config.Definitions)
                {
                    try
                    {
                        Dictionary<string, string> conditionsToCheck = ParseCondition(definition);

                        var attributesToLoad = new List<string>();
                        attributesToLoad.Add(definition.LabelAttributeKey);
                        foreach (var cond in conditionsToCheck)
                        {
                            attributesToLoad.Add(cond.Key);
                        }

                        var attributes = hook.BuilderAttributeAccessor.GetAttributes(element, new HashSet<string>(attributesToLoad));

                        if (!CheckConditions(definition, attributes, conditionsToCheck))
                        {
                            continue;
                        }

                        var labelAttributeValue = attributes[definition.LabelAttributeKey];

                        ret.Add(new ObjectToProcess(element, definition, labelAttributeValue));
                    }
                    catch (Exception ex)
                    {
                        Debug.Print(" ");
                    }
                }
            }
            hook.BuilderContext.Log($"{Name}: Identified {ret.Count} objects.");
            return ret;
        }

        private static Dictionary<string, string> ParseCondition(Definition definition)
        {
            var conditionsToCheck = new Dictionary<string, string>();

            //fix format
            definition.Condition = definition.Condition.Replace("&amp;", "&");
            //check if multiple conditions
            if (definition.Condition.Contains("&"))
            {
                //multiple conditions
                var conditions = definition.Condition.Split('&');
                foreach (var item in conditions)
                {
                    if (item.Contains("="))
                    {
                        var tmp = item.Split('=');
                        conditionsToCheck.Add(tmp[0], tmp[1]);
                    }
                }
            }
            else
            {
                //single condition
                if (definition.Condition.Contains("="))
                {
                    var tmp = definition.Condition.Split('=');
                    conditionsToCheck.Add(tmp[0], tmp[1]);
                }
            }

            return conditionsToCheck;
        }

        private bool CheckConditions(Definition def, IDictionary<string, string> attributes, Dictionary<string, string> conditionsToCheck)
        {
            var result = true;
            foreach (var cond in conditionsToCheck)
            {
                if (attributes.TryGetValue(cond.Key, out var queryAttributeValue))
                {
                    if (attributes.TryGetValue(def.LabelAttributeKey, out var labelAttributeValue))
                    {
                        if (def.LabelAttributeValueContains)
                        {
                            bool contains = queryAttributeValue.IndexOf(cond.Value, StringComparison.OrdinalIgnoreCase) >= 0;
                            if (!contains)
                                result = false;
                        }
                        else
                        {
                            if (!string.Equals(queryAttributeValue, cond.Value, StringComparison.InvariantCultureIgnoreCase))
                                result = false;
                        }
                    }
                    else
                        result = false;
                }
                else
                    result = false;

                if (!result) break;
            }
            return result;
        }
        private void GenerateGeometry(IPreExportHook hook, List<ObjectToProcess> objectsToProcess)
        {
            hook.BuilderContext.Log($"{Name}: Generating geometry..");
            Parallel.ForEach(objectsToProcess, (item) =>
            {
                foreach (var graphic in hook.BuilderGraphicAccessor.GetGraphics(item.Element))
                {
                    if (graphic.AspectName.ToLowerInvariant() != "simple physical")
                    {
                        continue;
                    }

                    try
                    {
                        Utility util = new Utility(item.Definition, hook);

                        var primitives = hook.BuilderGraphicAccessor.GetPrimitives(graphic);

                        IList<IGraphicData> commentGeometry;

                        if (item.Definition.Condition.ToLower().Contains("name=pipe"))
                        {
                            commentGeometry = util.GetCommentGeometry(item.LabelAttributeValue, primitives, item.Definition.Color).ToList();
                        }
                        else
                        {
                            var bounds = hook.BuilderGraphicAccessor.GetBounds(graphic);
                            var size = hook.Helpers.MathHelper.BoundsMath.GetSize(bounds);

                            var volume = size.X * size.Y * size.Z;
                            if (volume < item.Definition.MinimumSize)
                            {
                                continue;
                            }

                            var allAxis = true;
                            if (item.Definition.isRoutingItem == 1)
                                allAxis = false;

                            commentGeometry = util.CreateGraphicData(item.LabelAttributeValue, bounds, item.Definition.Color, allAxis).ToList();
                        }

                        if (!commentGeometry.Any())
                        {
                            continue;
                        }
                        item.GraphicToSet = commentGeometry;
                    }
                    catch (Exception ex)
                    {
                        hook.BuilderContext.LogError(ex.Message, ex);
                    }
                }
            });
            hook.BuilderContext.Log($"{Name}: Generating geometry done.");
        }

        private void WriteGeometry(IPreExportHook hook, List<ObjectToProcess> objectsToProcess)
        {
            hook.BuilderContext.Log($"{Name}: Writing geometry..");
            foreach (var item in objectsToProcess)
            {
                if (item.GraphicToSet != null)
                {
                    var newGraphic = hook.BuilderGraphicAccessor.CreateGraphic(item.Element, item.Definition.AspectName, item.GraphicToSet);
                    if (_config.Debug)
                    {
                        hook.BuilderContext.Log($"{Name}: Graphic added '{newGraphic.ParentElement.Uid}' '{newGraphic.AspectName}'");
                    }
                }
            }
            hook.BuilderContext.Log($"{Name}: Writing geometry done.");
        }
    }
}