﻿using System;
using System.Collections.Generic;
using System.Globalization;
using UniversalPlantViewerBuilderPlugin.Version1;
using System.Threading.Tasks;
using System.Linq;

namespace UPVBPluginGraphics
{
    public class Utility
    {
        private readonly Definition _config;
        private readonly IPreExportHook _hook;

        private readonly Vector3D[] _directions =
        {
            new Vector3D(0, 0, -Math.PI / 2),
            new Vector3D(0, -Math.PI / 2, -Math.PI / 2),
            new Vector3D( 0, Math.PI, -Math.PI / 2),
            new Vector3D(0, Math.PI / 2, -Math.PI / 2)
        };

        public Utility(Definition config, IPreExportHook hook)
        {
            _config = config;
            _hook = hook;
        }

        public IEnumerable<IGraphicData> GetCommentGeometry(string comment, IEnumerable<string> sourceGeometry, Color color)
        {
            var result = new List<IGraphicData>();
            var matrixHelper = _hook.Helpers.MathHelper.MatrixHelper;
            const double offset = 0.1f;
            var sourceList = sourceGeometry.ToList();
            Parallel.For(0, sourceList.Count, i =>
            {
                var row = sourceList[i];
                var cells = row.Split(' ');
                if (cells.Length > 2 && (cells[1] == "PlaceCylinder" || cells[1] == "PlaceRotationalSolid"))
                {
                    double cylinderLength;
                    double cylinderDiameter;
                    if (cells[1] == "PlaceCylinder")
                    {
                        cylinderLength = DoubleParse(cells[3]);
                        cylinderDiameter = DoubleParse(cells[2]);
                    }
                    else
                    {
                        cylinderLength = Math.Abs(DoubleParse(cells[10]));
                        cylinderDiameter = DoubleParse(cells[11]) * 2;
                    }
                    var cylinderRotation = new Vector3D(ToRad(DoubleParse(cells[4])), ToRad(DoubleParse(cells[5])), ToRad(DoubleParse(cells[6])));
                    var cylinderTranslation = new Vector3D(DoubleParse(cells[7]), DoubleParse(cells[8]), DoubleParse(cells[9]));
                    var textHeight = cylinderDiameter / 2 * _config.TextHeightFactor;
                    var textLength = GetTextLength(comment, textHeight);

                    if (textLength * 1.5 < cylinderLength)
                    {
                        var m = matrixHelper.GetTransformationMatrix(cylinderRotation, cylinderTranslation);
                        var mCylinderRotation = matrixHelper.GetTransformationMatrix(cylinderRotation, new Vector3D());

                        var whiteSpace = cylinderLength - textLength;
                        var radius = cylinderDiameter / 2;

                        if (textLength * 20 < cylinderLength)
                        {
                            var g = PlaceTextAtBeginningAndEndOfPipe(comment, color, matrixHelper, textHeight, m, mCylinderRotation, whiteSpace, radius);
                            lock (result)
                            {
                                result.AddRange(g);
                            }
                        }
                        else
                        {
                            var g = PlaceTextAtMiddleOfPipe(comment, color, matrixHelper, textHeight, m, mCylinderRotation, whiteSpace, radius);
                            lock (result)
                            {
                                result.AddRange(g);
                            }
                        }
                    }
                }
            });
            return result;
        }

        private List<IGraphicData> PlaceTextAtMiddleOfPipe(string comment, Color color, IMatrixMathHelper matrixHelper, double textHeight, Matrix m, Matrix mCylinderRotation, double whiteSpace, double radius)
        {
            var text = " " + comment + " ";
            var transformations = new[]
            {
                        new Transformation(_directions[0], new Vector3D(whiteSpace / 2, -radius, -textHeight / 2), text),
                        new Transformation(_directions[1], new Vector3D(whiteSpace / 2, -textHeight / 2, radius), text),
                        new Transformation(_directions[2], new Vector3D(whiteSpace / 2, radius, textHeight / 2), text),
                        new Transformation(_directions[3], new Vector3D(whiteSpace / 2, textHeight / 2, -radius), text)
                      };
            var g = FillDataResult(comment, transformations, matrixHelper, mCylinderRotation, m, textHeight, color);
            return g;
        }

        private List<IGraphicData> PlaceTextAtBeginningAndEndOfPipe(string comment, Color color, IMatrixMathHelper matrixHelper, double textHeight, Matrix m, Matrix mCylinderRotation, double whiteSpace, double radius)
        {
            var textLeadingWhiteSpaces = " " + comment;
            var textFollowingWhiteSpaces = comment + " ";

            var transformations = new[]
            {
                        new Transformation(_directions[0], new Vector3D(0, -radius, -textHeight / 2), textLeadingWhiteSpaces),
                        new Transformation(_directions[1], new Vector3D(0, -textHeight / 2, radius), textLeadingWhiteSpaces),
                        new Transformation(_directions[2], new Vector3D( 0, radius, textHeight / 2), textLeadingWhiteSpaces),
                        new Transformation(_directions[3], new Vector3D(0, textHeight / 2, -radius), textLeadingWhiteSpaces),

                        new Transformation(_directions[0], new Vector3D(whiteSpace, -radius, -textHeight / 2), textFollowingWhiteSpaces),
                        new Transformation(_directions[1], new Vector3D(whiteSpace, -textHeight / 2, radius), textFollowingWhiteSpaces),
                        new Transformation(_directions[2], new Vector3D(whiteSpace, radius, textHeight / 2), textFollowingWhiteSpaces),
                        new Transformation(_directions[3], new Vector3D(whiteSpace, textHeight / 2, -radius), textFollowingWhiteSpaces)
                      };
            var g = FillDataResult(comment, transformations, matrixHelper, mCylinderRotation, m, textHeight, color);
            return g;
        }

        private List<IGraphicData> FillDataResult(string comment, Transformation[] transformations, IMatrixMathHelper matrixHelper, Matrix mCylinderRotation, Matrix m, double textHeight, Color color)
        {
            var ret = new List<IGraphicData>();
            foreach (var transformation in transformations)
            {
                var angles = GetAngles(matrixHelper, mCylinderRotation, transformation.Rotation);
                var translation = matrixHelper.MultiplyPoint(m, transformation.Translation);
                var graphicData = CreateTextGraphicData(transformation.Comment, 0, textHeight, angles, translation, color);
                ret.Add(graphicData);
            }
            return ret;
        }

        private double GetTextLength(string comment, double textHeight)
        {
            var testText = _hook.BuilderGraphicAccessor.PlaceExtrudedText(comment + "__", 0, textHeight, new Vector3D(), new Vector3D());
            return _hook.Helpers.MathHelper.BoundsMath.GetSize(_hook.BuilderGraphicAccessor.GetBounds(testText)).Y;
        }

        private Vector3D GetAngles(IMatrixMathHelper matrixHelper, Matrix mCylinderRotation, Vector3D rotation)
        {
            var m = matrixHelper.GetTransformationMatrix(rotation, new Vector3D());
            var rot = matrixHelper.Multiply(mCylinderRotation, m);
            return matrixHelper.GetAngles(rot);
        }

        private IGraphicData CreateTextGraphicData(string text, double length, double height, Vector3D rotation, Vector3D translation, Color color)
        {
            return _hook.BuilderGraphicAccessor.PlaceExtrudedText(text, length, height, rotation, translation, new UniversalPlantViewerBuilderPlugin.Version1.Color(color.R, color.G, color.B, color.A));
        }

        public IEnumerable<IGraphicData> CreateGraphicData(string comment, Bounds3D boundingBox, Color color, bool allAxis)
        {
            const double offset = 0.1;
            var result = new List<IGraphicData>();

            var size = _hook.Helpers.MathHelper.BoundsMath.GetSize(boundingBox);
            var min = _hook.Helpers.MathHelper.BoundsMath.GetMin(boundingBox);
            var max = _hook.Helpers.MathHelper.BoundsMath.GetMax(boundingBox);

            var lengthX = size.X;
            var lengthY = size.Y;
            var textHeight = lengthX > lengthY ? lengthY * _config.TextHeightFactor : lengthX * _config.TextHeightFactor;
            var textLength = GetTextLength(comment, textHeight);

            List<Transformation> transformations = new List<Transformation>();
            if (allAxis)
            {
                transformations.Add(new Transformation(new Vector3D(0, 0, -Math.PI / 2), new Vector3D(boundingBox.Center.X - textLength / 2, min.Y - offset, boundingBox.Center.Z - offset / 2), comment));
                transformations.Add(new Transformation(new Vector3D(0, 0, Math.PI / 2), new Vector3D(boundingBox.Center.X + textLength / 2, max.Y + offset, boundingBox.Center.Z - offset / 2), comment));
                transformations.Add(new Transformation(new Vector3D(), new Vector3D(max.X + offset, boundingBox.Center.Y - textLength / 2, boundingBox.Center.Z - offset / 2), comment));
                transformations.Add(new Transformation(new Vector3D(0, 0, Math.PI), new Vector3D(min.X - offset, boundingBox.Center.Y + textLength / 2, boundingBox.Center.Z - offset / 2), comment));
            }
            else
            {
                if (lengthX > lengthY)
                {
                    //only add label on the x axis
                    transformations.Add(new Transformation(new Vector3D(0, 0, -Math.PI / 2), new Vector3D(boundingBox.Center.X - textLength / 2, min.Y - offset, boundingBox.Center.Z - offset / 2), comment));
                    transformations.Add(new Transformation(new Vector3D(0, 0, Math.PI / 2), new Vector3D(boundingBox.Center.X + textLength / 2, max.Y + offset, boundingBox.Center.Z - offset / 2), comment));
                }
                else
                {
                    //only add label on the y axis
                    transformations.Add(new Transformation(new Vector3D(), new Vector3D(max.X + offset, boundingBox.Center.Y - textLength / 2, boundingBox.Center.Z - offset / 2), comment));
                    transformations.Add(new Transformation(new Vector3D(0, 0, Math.PI), new Vector3D(min.X - offset, boundingBox.Center.Y + textLength / 2, boundingBox.Center.Z - offset / 2), comment));
                }
            }

            foreach (var transformation in transformations)
            {
                var graphicData = CreateTextGraphicData(transformation.Comment, 0, textHeight, transformation.Rotation, transformation.Translation, color);
                result.Add(graphicData);
            }

            return result;
        }

        private double DoubleParse(string s)
        {
            if (!double.TryParse(s, NumberStyles.Any, CultureInfo.InvariantCulture, out var result))
            {
                return double.NaN;
            }
            return result;
        }

        private static double ToRad(double degrees)
        {
            return degrees * (Math.PI / 180);
        }

        private struct Transformation
        {
            public readonly Vector3D Rotation;
            public readonly Vector3D Translation;
            public readonly string Comment;

            public Transformation(Vector3D rotation, Vector3D translation, string comment)
            {
                Rotation = rotation;
                Translation = translation;
                Comment = comment;
            }
        }
    }
}